// 2-norm distance between two units
function udist (a, b) {
  return Math.sqrt((a.x() - b.x())**2 + (a.y() - b.y())**2);
}

function vadd (u, v) {
  return [u[0] + v[0], u[1] + v[1]];
}

function vsub (u, v) {
  return [u[0] - v[0], u[1] - v[1]];
}

function vscale (k, v) {
  return [k * v[0], k * v[1]];
}

function vnorm (v) {
  return Math.sqrt(v[0]**2 + v[1]**2);
}

function dist (u, v) {
  return vnorm(vsub(u, v));
}

function vnormalize (v) {
  return vscale(1/vnorm(v), v);
}

function vdot (u, v) {
  return u[0]*v[0] + u[1]*v[1];
}

function rot90 (v) {
  return [-v[1], v[0]];
}

function rot (v, theta) {
  let cos = Math.cos(theta);
  let sin = Math.sin(theta);
  return [v[0]*cos - v[1]*sin, v[0]*sin + v[1]*cos];
}

// used by sim_battle to iterate over all units in the correct order
function compare_units (a, b) {
  if (a.isplayer && !b.isplayer) return -1;
  if (!a.isplayer && b.isplayer) return 1;
  return a._target_dist - b._target_dist;
}

// checks whether a line segment intersects a circle
function intersects (p1, p2, c, radius) {
  if (Math.max(vnorm(vsub(p1, c)), vnorm(vsub(p2, c))) < radius) return true;
  let dir = vnormalize(vsub(p2, p1));
  let how_far_ahead = vdot(dir, vsub(c, p1));
  if (how_far_ahead < 0 || how_far_ahead > vdot(dir, vsub(p2, p1))) return false;
  let normal = rot90(dir);
  return Math.abs(vdot(normal, vsub(c, p1))) <= radius;
}

// generate a graph of circle tangents from the units
// TODO: currently O(n^3), might be possible to do more efficiently
// TODO: might want to merge nodes within epsilon of each other
function make_tangent_graph (units, radius) {
  units.forEach(function (u) { u._nodes = []; });
  // a key in graph is a node, which is a 2-tuple containing the unit index (ui
  // or vi), and the coordinates of the node. The value is an adjacency list,
  // each item of which is a 3-tuple containing the other node, the length of
  // the edge, and the direction required to reach that node
  let graph = {};
  for (var ui = 0; ui < units.length; ui++) {
    let u = units[ui];
    for (var vi = ui+1; vi < units.length; vi++) {
      let v = units[vi];
      // need to iterate over 4 different lines between circles of u and v

      // two of these are parallel to v.pos - u.pos; they stay on the same side of both circles
      let disp = vsub(v.pos, u.pos);
      let dir = normalize(disp);
      let normal = rot90(dir);
      let s1 = [vadd(u.pos, vscale(radius, normal)),
                vadd(v.pos, vscale(radius, normal))];
      let s2 = [vsub(u.pos, vscale(radius, normal)),
                vsub(v.pos, vscale(radius, normal))];

      // two of these cross between the circles; they are solutions to
      // (v.pos - u.pos) * x = radius, with ||x|| = 1
      let min_norm = 2 * radius / vnorm(disp);
      let offset = vscale(Math.sqrt(1 - min_norm**2), normal);
      let s3_normal = vadd(dir, offset);
      let s4_normal = vadd(dir, offset);
      let s3 = [vadd(u.pos, s3_normal), vsub(v.pos, s3_normal)];
      let s4 = [vsub(u.pos, s3_normal), vadd(v.pos, s3_normal)];

      for (const [p1, p2] of [s1, s2, s3, s4]) {
        if (!units.some(w => w !== u && w !== v && intersects(p1, p2, w.pos, radius))) {
          let length = vnorm(vsub(p1, p2));
          let n1 = [ui, p1];
          let n2 = [vi, p2];
          if (!(n1 in graph)) graph[n1] = [];
          if (!(n2 in graph)) graph[n2] = [];
          graph[n1].push([n2, length, vsub(p2, p1)]);
          graph[n2].push([n1, length, vsub(p1, p2)]);
        }
      }
    }
  }

}

function sim_battle () {
  // naive pathfinding algorithm: iterate over player units first, and iterate
  // in order of unit -> target distance. try to move directly toward target at
  // full speed, the adjust to handle collisions
  let units = st.battle.aunits;
  if (!units.some(u => u.isplayer)) {
    st.battle.winner = 'enemy';
    return;
  } else if (!units.some(u => !u.isplayer)) {
    st.battle.winner = 'player';
    return;
  }

  for (const au of units) {
    let foes = units.filter(u => u.isplayer !== au.isplayer);
    if (au._target) {
      au._target_dist = udist(au, au._target);
      if (au._target.isalive && au.can_hit(au._target)) continue;
    }
    au._target = null;
    au._target_dist = Infinity;
    for (const foe of foes) {
      let dist = udist(au, foe);
      if (dist < au._target_dist) {
        au._target = foe;
        au._target_dist = dist;
      }
    }
  }
  units.sort(compare_units);
  st._units = units; // for debug purposes

  for (const au of units) {
    if (au.can_hit(au._target)) continue;
    let maxmv = au.unit.stats.spd / st.fps;
    let goal = vsub(au._target.pos, au.pos);
    let dir = vnormalize(goal);
    au.pos = vadd(au.pos, vscale(maxmv, dir));
    for (var ui = 0; ui < units.length; ui++) {
      let u = units[ui];
      if (u === au) continue;
      let intrusion = u.radius + au.radius - udist(u, au);
      if (intrusion > 0) {
        au.pos = vadd(au.pos, vscale(intrusion+st.epsilon, vnormalize(vsub(au.pos, u.pos))));
        ui = -1;
        continue;
      }
    }
  }

  for (const effect of st.effects) {
    if (effect.tick) effect.tick(effect);
    if (st.frame >= effect.timeout) {
      effect._delete = true;
    }
  }
  st.effects = st.effects.filter(e => !e._delete);

  for (const au of units) {
    au.tick();
  }
}

function aa_dmg (src, target) {
  return 10 * src.unit.stats.atk / target.unit.stats.def;
}
